#!/usr/bin/env -S python -u

import argparse
import atexit
import logging
import logging.handlers
import multiprocessing as mp
import pathlib

from sats_receiver import HOMEDIR, LOGSDIR, TLEDIR, RECDIR
from sats_receiver.async_signal import AsyncSignal
from sats_receiver.manager import ReceiverManager
from sats_receiver.utils import SysUsage


def setup_logging(q: mp.Queue, log_lvl: int):
    if not isinstance(log_lvl, int):
        raise ValueError('Invalid log level: %s' % log_lvl)

    logger = logging.getLogger()
    logger.setLevel(log_lvl)
    logger.addHandler(logging.handlers.QueueHandler(q))
    mp.get_logger().setLevel(log_lvl)

    fmt = logging.Formatter('%(asctime)s %(levelname)s: %(name)s: %(message)s')
    sh = logging.StreamHandler()
    sh.setFormatter(fmt)
    fh = logging.handlers.TimedRotatingFileHandler(LOGSDIR / 'sats_receiver.log', 'midnight')
    fh.setFormatter(fmt)

    qhl = logging.handlers.QueueListener(q, sh, fh)
    qhl.start()
    atexit.register(qhl.stop)

    # PIL logging level
    pil_logger = logging.getLogger('PIL')
    pil_logger.setLevel(logging.DEBUG + 2)


if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    ap.add_argument('config', type=pathlib.Path, help='Config file path')
    ap.add_argument('--log', default='INFO', type=(lambda x: getattr(logging, x.upper(), None)),
                    help='Logging level, INFO default')
    ap.add_argument('--sysu', default=SysUsage.DEFAULT_INTV, type=int,
                    help='System Usages info timeout in seconds, 1 hour default')
    args = ap.parse_args()

    for d in LOGSDIR, TLEDIR, RECDIR:
        if not d.exists():
            d.mkdir(parents=True, exist_ok=True)

    q = mp.Queue()
    setup_logging(q, args.log)

    logging.info('Hello!')

    asig = AsyncSignal(['SIGABRT', 'SIGHUP', 'SIGINT', 'SIGTERM', 'SIGUSR1', 'SIGUSR2', 'SIGBREAK'])
    mng = ReceiverManager(q, args.config, args.sysu)

    while not mng.action():
        signame = asig.wait(1)
        if signame:
            if 'USR' in signame:
                # TODO
                pass
            else:
                mng.stop()
                logging.info('Exit by %s', signame)
                break

    mng.wait()
    logging.info('Bye!')



import datetime as dt
import pathlib
import time
import threading
import queue

from typing import Mapping
from test import support
from unittest import TestCase

import numpy as np

from sats_receiver import utils


HERE = pathlib.Path(__file__).parent
TIMEOUT = support.SHORT_TIMEOUT


class SchedThread(threading.Thread):
    def __init__(self, sch: utils.Scheduler):
        super().__init__()
        self._q = queue.SimpleQueue()
        self.sch = sch

    def run(self):
        while 1:
            try:
                self._q.get_nowait()
            except queue.Empty:
                pass
            else:
                break

            d: dt.timedelta = self.sch.action()
            if d:
                time.sleep(d.seconds)

    def stop(self):
        self._q.put_nowait(1)


class TestScheduler(TestCase):
    """
    Test case examples are taken from the python repository
    https://github.com/python/cpython/blob/main/Lib/test/test_sched.py
    """

    def setUp(self):
        self._sch = utils.Scheduler()
        self._now = dt.datetime.now(dt.timezone.utc)
        self._l = []

    def test_plan(self):
        for x in [5, 4, 3, 2, 1]:
            self._sch.plan(self._now + dt.timedelta(seconds=x / 100), self._l.append, x, prior=1)
        time.sleep(0.2)
        self._sch.action()

        self.assertEqual([1, 2, 3, 4, 5], self._l)

    def test_plan_delay(self):
        self._sch.plan(self._now + dt.timedelta(seconds=1), self._l.append, 1, prior=1)
        delay = self._sch.action()

        self.assertGreaterEqual(dt.timedelta(seconds=1), delay)
        self.assertLessEqual(dt.timedelta(seconds=0), delay)

    def test_plan_concurrent(self):
        q = queue.SimpleQueue()
        fn = q.put
        self._sch.plan(self._now + dt.timedelta(seconds=0.1), fn, 1)
        self._sch.plan(self._now + dt.timedelta(seconds=0.3), fn, 3)
        t = SchedThread(self._sch)
        t.start()

        try:
            self.assertEqual(1, q.get(timeout=TIMEOUT))
            self.assertTrue(q.empty())

            now = dt.datetime.now(dt.timezone.utc)
            for x in [4, 5, 2]:
                self._sch.plan(now + dt.timedelta(seconds=(x - 1) / 10), fn, x)

            self.assertEqual(2, q.get(timeout=TIMEOUT))
            self.assertEqual(3, q.get(timeout=TIMEOUT))
            self.assertTrue(q.empty())

            self.assertEqual(4, q.get(timeout=TIMEOUT))
            self.assertTrue(q.empty())

            self.assertEqual(5, q.get(timeout=TIMEOUT))
            self.assertTrue(q.empty())

            time.sleep(1)
        finally:
            t.stop()
            t.join()

        self.assertTrue(q.empty())

    def test_priority(self):
        cases = [
            ([1, 2, 3, 4, 5], [1, 2, 3, 4, 5]),
            ([1, 2, 3, 4, 5], [5, 4, 3, 2, 1]),
            ([1, 2, 3, 4, 5], [2, 5, 3, 1, 4]),
            ([1, 1, 2, 2, 3], [1, 2, 3, 2, 1]),
        ]
        for expected, priorities in cases:
            with self.subTest(expected=expected, priorities=priorities):
                for priority in priorities:
                    self._sch.plan(self._now, self._l.append, priority, prior=priority)
                self._sch.action()
                self.assertEqual(expected, self._l)

                self.assertTrue(self._sch.empty())
                self._l.clear()

    def test_cancel(self):
        event1 = self._sch.plan(self._now + dt.timedelta(seconds=0.01), self._l.append, 1)
        event2 = self._sch.plan(self._now + dt.timedelta(seconds=0.02), self._l.append, 2)
        event3 = self._sch.plan(self._now + dt.timedelta(seconds=0.03), self._l.append, 3)
        event4 = self._sch.plan(self._now + dt.timedelta(seconds=0.04), self._l.append, 4)
        event5 = self._sch.plan(self._now + dt.timedelta(seconds=0.05), self._l.append, 5)
        self._sch.cancel(event1, event5)
        time.sleep(0.2)
        self._sch.action()

        self.assertEqual([2, 3, 4], self._l)

    def test_cancel_same_time(self):
        self._sch.plan(self._now, self._l.append, 'a')
        b = self._sch.plan(self._now, self._l.append, 'b')
        self._sch.plan(self._now, self._l.append, 'c')
        self._sch.cancel(b)
        self._sch.action()

        self.assertEqual(['a', 'c'], self._l)

    def test_cancel_expired_evt(self):
        event1 = self._sch.plan(self._now, self._l.append, 1)
        event2 = self._sch.plan(self._now, self._l.append, 2)
        self._sch.action()
        self._sch.cancel(event1, event2)

        self.assertTrue(self._sch.empty())
        self.assertEqual([1, 2], self._l)

    def test_cancel_invalid_evt(self):
        self._sch.plan(self._now, self._l.append, 1)
        self._sch.plan(self._now, self._l.append, 2)
        self._sch.cancel()
        self._sch.cancel(1)
        self._sch.cancel(2.0, '3', None)
        self.assertFalse(self._sch.empty())

    def test_cancel_concurrent(self):
        q = queue.SimpleQueue()
        fn = q.put
        event1 = self._sch.plan(self._now + dt.timedelta(seconds=0.1), fn, 1)
        event2 = self._sch.plan(self._now + dt.timedelta(seconds=0.2), fn, 2)
        event4 = self._sch.plan(self._now + dt.timedelta(seconds=0.4), fn, 4)
        event5 = self._sch.plan(self._now + dt.timedelta(seconds=0.5), fn, 5)
        event3 = self._sch.plan(self._now + dt.timedelta(seconds=0.3), fn, 3)
        t = SchedThread(self._sch)
        t.start()

        try:
            self.assertEqual(1, q.get(timeout=TIMEOUT))
            self.assertTrue(q.empty())

            self._sch.cancel(event2, event5)
            self.assertTrue(q.empty())

            self.assertEqual(3, q.get(timeout=TIMEOUT))
            self.assertTrue(q.empty())

            self.assertEqual(4, q.get(timeout=TIMEOUT))
            self.assertTrue(q.empty())

            time.sleep(1)
        finally:
            t.stop()
            t.join()

        self.assertTrue(q.empty())

    def test_empty(self):
        self.assertTrue(self._sch.empty())

        for x in [0.05, 0.04, 0.03, 0.02, 0.01]:
            self._sch.plan(self._now + dt.timedelta(seconds=x), self._l.append, x)

        self.assertFalse(self._sch.empty())

        time.sleep(0.2)
        self._sch.action()

        self.assertTrue(self._sch.empty())

    def test_clear(self):
        for x in [0.05, 0.04, 0.03, 0.02, 0.01]:
            self._sch.plan(self._now + dt.timedelta(seconds=x), self._l.append, x)
        self._sch.clear()

        self.assertTrue(self._sch.empty())

    def test_args_kwargs(self):
        def fn(*a, **b):
            self._l.append((a, b))

        self._sch.plan(self._now, fn)
        self._sch.plan(self._now, fn, 1, 2)
        self._sch.plan(self._now, fn, 'a', 'b')
        self._sch.plan(self._now, fn, 1, 2, foo=3)
        self._sch.action()
        self.assertCountEqual([
            ((), {}),
            ((1, 2), {}),
            (('a', 'b'), {}),
            ((1, 2), {'foo': 3})
        ], self._l)

    def test_plan_time(self):
        def fn():
            self._l.append(dt.datetime.now(dt.timezone.utc))

        t = self._now + dt.timedelta(seconds=1)
        self._sch.plan(t, fn)
        time.sleep(1)
        self._sch.action()

        self.assertLess(t, self._l[0])
        self.assertGreater(t + dt.timedelta(seconds=1), self._l[0])


class TestUtils(TestCase):
    def test_sysu_collect_debug_log(self):
        sysu = utils.SysUsage('Test', 0)
        with self.assertLogs(level='DEBUG'):
            sysu.collect()

    def test_numbi_disp_ok(self):
        cases = (
            ('0b', -1),
            ('0b', 0),
            ('5b', 'hello'),
            ('5b', b'world'),
            ('3b', [1, 2, 3]),
            ('1b', 1.9),
            ('999b', 999),
            ('.99K', 1023),
            ('1.0K', 1024),
            ('4.5K', (4 << 10) + 512),
            ('1.0M', 1 << 20),
            ('2.0G', 2 << 30),
            ('3.0T', 3 << 40),
            ('4.0P', 4 << 50),
            ('5.0E', 5 << 60),
            ('6.0Z', 6 << 70),
            ('7.0Y', 7 << 80),
            (12, -1, 12),
            ('-12', None, '-12'),
            (0, 0, 0),
        )
        for i in cases:
            with self.subTest(expect=i[0], num=i[1]):
                self.assertEqual(i[0], utils.numbi_disp(*i[1:]))

    def test_numbi_disp_type_error(self):
        with self.assertRaises(TypeError):
            utils.numbi_disp(self)

    def test_num_disp(self):
        cases = (
            ('-1', -1),
            ('0', 0),
            ('1.9', 1.9),
            ('999', 999),
            ('1k', 1005),
            ('1.02k', 1024),
            ('4.5k', 4 * 10e2 + 500),
            ('1M', 1 * 10e5),
            ('2G', 2 * 10e8),
            ('3T', 3 * 10e11),
            ('4P', 4 * 10e14),
            ('5E', 5 * 10e17),
            ('6Z', 6 * 10e20),
            ('7Y', 7 * 10e23),
            ('-1.2345k', -1234.5, 4),
            ('-1.2345k', -1234.5, 6),
        )
        for i in cases:
            with self.subTest(expect=i[0], num=i[1]):
                self.assertEqual(i[0], utils.num_disp(*i[1:]))

    def test_sec(self):
        cases = (
            ('855s', 855),
            ('12s', 12.3456, 0),
            ('12.3s', 12.3456, 1),
            ('12.35s', 12.3456, 2),
            ('12.346s', 12.3456, 3),
            ('12.3456s', 12.3456, 4),
            ('12.3456s', 12.3456, 6),
        )
        for i in cases:
            with self.subTest(expect=i[0], sec=i[1]):
                self.assertEqual(i[0], utils.sec(*i[1:]))

    def test_doppler_shift(self):
        cases = (
            (1000000, -1000),
            (1000000, -1),
            (1000000, 0),
            (1000000, 1),
            (1000000, 1000),
            (1, 10),
        )
        for freq, vel in cases:
            with self.subTest(freq=freq, vel=vel):
                if vel > 0:
                    a = self.assertGreater
                elif vel < 0:
                    a = self.assertLess
                else:
                    a = self.assertEqual

                a(freq, utils.doppler_shift(freq, vel))

    def test_azimuth(self):
        cases = (
            (np.radians(0), np.radians((0, 0)), np.radians((0, 90))),
            (np.radians(0), np.radians((0, -90)), np.radians((0, 90))),
            (np.radians(90), np.radians((0, 0)), np.radians((90, 0))),
            (np.radians(180), np.radians((0, 0)), np.radians((0, -90))),
            (np.radians(180), np.radians((0, 90)), np.radians((0, -90))),
            (np.radians(-90), np.radians((0, 0)), np.radians((-90, 0))),
            (np.radians(-90), np.radians((90, 0)), np.radians((0, 0))),
        )
        for expected, a_lonlat, b_lonlat in cases:
            with self.subTest(expected=expected, a_lonlat=a_lonlat, b_lonlat=b_lonlat):
                self.assertEqual(expected, utils.azimuth(a_lonlat, b_lonlat))

    def test_map_shapes_ok(self):
        color = 'coral', '#ff7f50', (0xff, 0x7f, 0x50), (0xff, 0x7f, 0x50, 255)

        cfg = dict(
            shapes_dir=HERE / 'files',
            line_width=1,
            shapes=(
                (1, 'ne_110m_graticules_30.zip', color[0]),
            ),
            points=dict(
                observer=dict(color=color[1], type='+', size=[3, 21]),
                another=dict(color=color[2], type='o', size=12, lonlat=[0, 0]),
                other=dict(color=color[3], type='+', size=[3, 21], lonlat=[81, -33]),
            ),
        )
        ms = utils.MapShapes(cfg)

        self.assertEqual(HERE / 'files', ms.shapes_dir)
        self.assertEqual(1, ms.line_width)

        for points, col in ms.iter():
            x = isinstance(points, Mapping)

            with self.subTest(points=points if x else 'shapefile', color=col):
                self.assertTupleEqual(color[3], col)

                if x:
                    if points['name'] != 'observer':
                        self.assertIn('lonlat', points)
                    if points['type'] == '+':
                        self.assertEqual(2, len(points['size']))
                    elif points['type'] == 'o':
                        self.assertIsInstance(points['size'], (int, float))

    def test_map_shapes_fail(self):
        cases = (
            (dict(points=dict(xxx=dict(type='+'))), KeyError),
            (dict(points=dict(xxx=dict(type='+', lonlat=[0, 0]))), KeyError),
            (dict(points=dict(xxx=dict(type='x', lonlat=[0, 0], color=None))), TypeError),
            (dict(points=dict(xxx=dict(type='x', lonlat=[0, 0], color='black'))), ValueError),
            (dict(points=dict(xxx=dict(type='+', lonlat=[0, 0], color=0))), KeyError),
            (dict(points=dict(xxx=dict(type='+', lonlat=[0, 0], color=0, size=None))), TypeError),
            (dict(points=dict(xxx=dict(type='+', lonlat=[0, 0], color=0, size=[None]))), AssertionError),
            (dict(points=dict(xxx=dict(type='o', lonlat=[0, 0], color=0, size=[None]))), TypeError),
        )
        for cfg, exc in cases:
            with self.subTest(cfg=cfg):
                with self.assertRaises(exc):
                    utils.MapShapes(cfg)



import datetime as dt
import enum
import logging
import math
import os
import pathlib

from typing import Union

import dateutil.tz
import ephem
import numpy as np
import scipy as sp
import scipy.signal

from PIL import Image, ImageDraw

from sats_receiver import utils


class AptChannel(enum.Enum):
    A = enum.auto()
    B = enum.auto()


class AptWedgeNum(enum.IntEnum):
    ONE = 7
    ZERO = 8
    PRT_1 = 9
    PRT_2 = 10
    PRT_3 = 11
    PRT_4 = 12
    PATCH_TEMP = 13
    BACK_SCAN = 14
    CHAN_IDENT = 15


class AptChannelName(enum.Enum):
    _1 = '1'    # Visible (0.58 – 0.68 μm)
    _2 = '2'    # Near Infrared (0.325 – 1.10 μm)
    _3a = '3a'  # Thermal Infrared (1.58 – 1.64 μm)
    _4 = '4'    # Thermal Infrared (10.30 – 11.30 μm)
    _5 = '5'    # Thermal Infrared (11.50 – 12.50 μm)
    _3b = '3b'  # Thermal Infrared (3.55 – 3.93 μm)


class AptTelemetry:
    CHANNELS_NAMES = [*AptChannelName, None, None, None]

    def __init__(self, ch_a: np.ndarray, ch_b: np.ndarray):
        self.ch_a = ch_a
        self.ch_b = ch_b

        self.channel_a_name = self._channel_name(AptChannel.A)
        self.channel_b_name = self._channel_name(AptChannel.B)

    def wedge_value(self, num: AptWedgeNum, chan: AptChannel = None) -> float:
        if chan is None:
            return (self.ch_a[num] + self.ch_b[num]) / 2
        if chan == AptChannel.A:
            return self.ch_a[num]
        if chan == AptChannel.B:
            return self.ch_b[num]

    def _channel_name(self, chan: AptChannel) -> AptChannelName:
        return min(zip(self.CHANNELS_NAMES, self.ch_a[:9]),
                   key=lambda p: abs(p[1] - self.wedge_value(AptWedgeNum.CHAN_IDENT, chan)))[0]


class AptFrame:
    def __init__(self, data: np.ndarray, tlm_a: np.ndarray, tlm_b: np.ndarray):
        self.data = data
        self.tlm = AptTelemetry(tlm_a, tlm_b)

    @property
    def img_a(self) -> np.ndarray:
        return self.data[:, Apt.IMAGE_A_START:Apt.TLM_A_START]

    @property
    def img_b(self) -> np.ndarray:
        return self.data[:, Apt.IMAGE_B_START:Apt.TLM_B_START]

    def set_tlm_calibration(self, c: np.ndarray):
        self.tlm.ch_a[:AptWedgeNum.PRT_1] = c
        self.tlm.ch_b[:AptWedgeNum.PRT_1] = c


class Apt:
    CARRIER_FREQ = 2400
    FRAME_WIDTH = 2080
    PIX_WIDTH = 4
    FINAL_RATE = 4160
    WORK_RATE = FINAL_RATE * PIX_WIDTH
    SAMPLES_PER_WORK_ROW = FRAME_WIDTH * PIX_WIDTH
    LINE_DUR = dt.timedelta(milliseconds=500)

    SYNC_WIDTH = 39
    SPACE_WIDTH = 47
    IMAGE_WIDTH = 909
    TLM_WIDTH = 45

    SYNC_A_START = 0
    SPACE_A_START = SYNC_A_START + SYNC_WIDTH
    IMAGE_A_START = SPACE_A_START + SPACE_WIDTH
    TLM_A_START = IMAGE_A_START + IMAGE_WIDTH

    SYNC_B_START = TLM_A_START + TLM_WIDTH
    SPACE_B_START = SYNC_B_START + SYNC_WIDTH
    IMAGE_B_START = SPACE_B_START + SPACE_WIDTH
    TLM_B_START = IMAGE_B_START + IMAGE_WIDTH

    BLOCK_HEIGHT = 8
    BLOCKS_NUM = 16
    FRAME_HEIGHT = BLOCK_HEIGHT * BLOCKS_NUM

    SYNC_A = np.array([*map(float, '000011001100110011001100110011000000000')],
                      dtype=np.float32).repeat(PIX_WIDTH) * 2 - 1
    SYNC_B = np.array([*map(float, '000011100111001110011100111001110011100')],
                      dtype=np.float32).repeat(PIX_WIDTH) * 2 - 1
    WEDGE_SAMPLE = np.array([*np.interp([31, 63, 95, 127, 159, 191, 223, 255, 0], [0, 255], [-1, 1]),
                             0, 0, 0, 0, 0, 0, 0], dtype=np.float32).repeat(BLOCK_HEIGHT)

    # TODO: anoter values???
    NOAA_AVHRR_FOV = 110.74

    @classmethod
    def from_apt(cls, aptf: pathlib.Path) -> 'Apt':
        """
        APT file format:
            * 3 lines of TLE data
            * observer lonlat, 2 x double, degrees
            * end time UTC timestamp, double
            * data
        """

        with aptf.open('rb') as f:
            sat_name = f.readline().decode('ascii').rstrip()
            l1 = f.readline().decode('ascii').rstrip()
            l2 = f.readline().decode('ascii').rstrip()
            sat_tle = sat_name, l1, l2
            observer_lonlat = [*np.frombuffer(f.read(np.dtype(np.double).itemsize * 2), dtype=np.double)]
            end_time = np.frombuffer(f.read(np.dtype(np.double).itemsize), dtype=np.double)[0]
            data = np.fromfile(f, dtype=np.float32)

        x = cls(sat_name, aptf, None, None, sat_tle, observer_lonlat)
        x.data = np.resize(data, (data.size // cls.FRAME_WIDTH, cls.FRAME_WIDTH))
        x.end_time = dt.datetime.fromtimestamp(end_time, dateutil.tz.tzutc())
        x.synced = True

        return x

    def __init__(self,
                 sat_name: str,
                 data_file: pathlib.Path,
                 corr_file: Union[pathlib.Path, None],
                 peaks_file: Union[pathlib.Path, None],
                 sat_tle: tuple[str, str, str],
                 observer_lonlat: tuple[float, float]):
        self.prefix = f'{self.__class__.__name__}: {sat_name}'
        self.log = logging.getLogger(self.prefix)

        self.sat_name = sat_name
        self.data_file = data_file
        self.corr_file = corr_file
        self.peaks_file = peaks_file
        self.peak_coef = 0.4
        self.dev = 0.02
        self.synced = False
        self.sat_ephem = ephem.readtle(*sat_tle)
        self.sat_tle = sat_tle
        self.observer_lonlat = observer_lonlat

        self.dist_dev = int(self.SAMPLES_PER_WORK_ROW * self.dev)
        self.dist_min = self.SAMPLES_PER_WORK_ROW - self.dist_dev
        self.dist_max = self.SAMPLES_PER_WORK_ROW + self.dist_dev

        t = 0
        if self.data_file and self.data_file.is_file():
            t = self.data_file.stat().st_mtime
        self.end_time = dt.datetime.fromtimestamp(t, dateutil.tz.tzutc())
        self.data = np.empty((0, self.FRAME_WIDTH))
        self.map_overlay = np.empty((0, self.IMAGE_WIDTH, 4))

    def to_apt(self, out_dir: pathlib.Path) -> tuple[pathlib.Path, int]:
        """
        APT file format:
            * 3 lines of TLE data
            * observer lonlat, 2 x double, degrees
            * end time UTC timestamp, double
            * data
        :return: result file path and size
        """

        res_fn = out_dir / self.end_time.strftime(f'{self.sat_name}_%Y-%m-%d_%H-%M-%S,%f.apt')
        sz = 0

        with res_fn.open('wb') as f:
            sz += f.write('\n'.join((*self.sat_tle, '')).encode('ascii'))
            sz += f.write(np.array(self.observer_lonlat, dtype=np.double).tobytes())
            sz += f.write(np.array(self.end_time.timestamp(), dtype=np.double).tobytes())
            sz += f.write(self.data.tobytes())

        os.utime(res_fn, (self.end_time.timestamp(), self.end_time.timestamp()))

        return res_fn, sz

    def process(self):
        """
        Run apr data process
        :return: True if process terminated with error
        """
        self.log.debug('process...')

        if not self.synced:
            try:
                tail_cutted, data, peaks_idx = self._prepare_data()
                if not data.size or peaks_idx.size < 5:
                    raise IndexError
            except IndexError:
                self.log.error('invalid received data')
                return 1

            self._syncing(tail_cutted, data, peaks_idx)

        return self._read_telemetry()

    def _prepare_data(self) -> tuple[int, np.ndarray, np.ndarray]:
        data: np.ndarray = np.fromfile(self.data_file, dtype=np.float32)
        corrs: np.ndarray = np.fromfile(self.corr_file, dtype=np.float32)
        peaks: np.ndarray = np.fromfile(self.peaks_file, dtype=np.byte)

        x = np.flatnonzero(corrs > (np.max(corrs[np.flatnonzero(peaks)]) * self.peak_coef))
        start_pos, end_pos = x[0], x[-1]
        tail_cutted = data.size - end_pos

        return tail_cutted, data[start_pos:end_pos], np.flatnonzero(peaks[start_pos:end_pos])

    def _syncing(self, tail_cutted, data, peaks_idx):
        self.log.debug('syncing...')

        peaks = [peaks_idx[0]]
        it = iter(range(1, peaks_idx.size))
        for i in it:
            prev = peaks_idx[i - 1]
            cur = peaks_idx[i]
            d = cur - prev

            if d < self.dist_min or d > self.dist_max:
                i_ = i
                for i_ in it:
                    prev_ = peaks_idx[i_ - 1]
                    cur_ = peaks_idx[i_]
                    k_ = prev_ - prev
                    d_ = cur_ - prev_

                    if self.dist_min < d_ < self.dist_max:
                        for j in range(1, round(k_ / self.SAMPLES_PER_WORK_ROW)):
                            peaks.append(prev + self.SAMPLES_PER_WORK_ROW * j)

                        k_ = prev_ - peaks[-1]
                        if self.dist_min < k_ < self.dist_max or k_ > self.dist_dev:
                            peaks.append(prev_)
                        else:
                            peaks[-1] = prev_

                        peaks.append(cur_)
                        break

                else:
                    prev_ = peaks_idx[i_ - 1]
                    cur_ = peaks_idx[i_]
                    d_ = cur_ - prev_

                    m = round(d_ / self.SAMPLES_PER_WORK_ROW)
                    if self.dist_min * m < d_ < self.dist_max * m:
                        m += 1

                    for j in range(1, m):
                        peaks.append(prev_ + self.SAMPLES_PER_WORK_ROW * j)

            else:
                peaks.append(cur)

        result = np.full((len(peaks), self.SAMPLES_PER_WORK_ROW), np.nan, dtype=np.float32)
        without_last = err = 0

        for idx, i in enumerate(peaks):
            try:
                x = data[i:peaks[idx + 1]]
            except IndexError:
                z = data.size - i
                if z < self.dist_min:
                    without_last = 1
                    break

                x = data[i:i + self.SAMPLES_PER_WORK_ROW]

            try:
                x = sp.signal.resample(x, self.SAMPLES_PER_WORK_ROW)
            except ValueError as e:
                if not err:
                    self.log.debug('error on line resample: %s', e)
                    err = 1
                continue

            result[idx] = x

        result_end_pos = (np.argmax(np.isnan(result).all(axis=1)) or result.shape[0]) - without_last
        result_tail_cutted = data.size / self.SAMPLES_PER_WORK_ROW - result_end_pos
        tail_cutted /= self.SAMPLES_PER_WORK_ROW

        self.end_time -= dt.timedelta(milliseconds=(tail_cutted + result_tail_cutted) * 500)
        self.data = result[0:result_end_pos, self.PIX_WIDTH // 2::self.PIX_WIDTH]
        self.synced = True

    def _read_telemetry(self):
        self.log.debug('read telemetry...')

        if self.data.shape[0] < self.WEDGE_SAMPLE.size:
            self.log.error('recording too short for telemetry decoding')
            return 1
        if self.data.shape[0] < self.WEDGE_SAMPLE.size * 2:
            self.log.warning('reading telemetry on short recording, expect unreliable results')

        tlm_a = self.data[0:, self.TLM_A_START:self.SYNC_B_START]
        tlm_b = self.data[0:, self.TLM_B_START:]

        # search the best template with minimum noise
        mean_a = np.mean(tlm_a, 1, dtype=np.float32)
        mean_b = np.mean(tlm_b, 1, dtype=np.float32)
        variance = (np.var(tlm_a, 1, dtype=np.float32)
                    + np.var(tlm_b, 1, dtype=np.float32)) / 2
        corr = np.correlate(mean_a, self.WEDGE_SAMPLE) + np.correlate(mean_b, self.WEDGE_SAMPLE)
        qual = np.array([v / np.std(variance[i:i + self.WEDGE_SAMPLE.size], dtype=np.float32)
                         for i, v in enumerate(corr)], dtype=np.float32)
        best = np.argmax(qual)

        # form the best telemetry data and create combine contrast values
        tlm_a = np.reshape(mean_a[best:best + self.FRAME_HEIGHT], (-1, self.BLOCK_HEIGHT)).mean(1, dtype=np.float32)
        tlm_b = np.reshape(mean_b[best:best + self.FRAME_HEIGHT], (-1, self.BLOCK_HEIGHT)).mean(1, dtype=np.float32)
        clb = (tlm_a[:9] + tlm_b[:9]) / 2
        hi, lo = clb[7:10]
        rng = hi - lo

        if rng > 0:
            # image correction by contrast values
            self.data = (self.data - lo) / rng
        else:
            self.log.warning('invalid telemetry data, perhaps is too noisy')

        # TODO
        # frames_num = math.ceil(mean_a.shape[0] / self.FRAME_HEIGHT)
        # first_frame_height = best % self.FRAME_HEIGHT
        #
        # frames format: start, height, quality
        # frames = [0, first_frame_height, qual[0]]
        # for i in range(first_frame_height, self.data.shape[0], self.FRAME_HEIGHT):
        #     try:
        #         q = qual[i]
        #     except IndexError:
        #         q = 0
        #     frames.append((i, self.FRAME_HEIGHT, q))
        # a, b, c = frames[-1]
        # frames[-1] = a, self.data.shape[0] - a, c

    def create_maps_overlay(self, shapes: utils.MapShapes):
        self.log.debug('create maps overlay...')

        ss_factor = 4
        self._height = height = self.data.shape[0]
        self._ss_height = height * ss_factor
        self._half_ss_height = self._ss_height / 2
        self._ss_width = self.IMAGE_WIDTH * ss_factor
        self._half_ss_width = self._ss_width / 2
        self._line_width = shapes.line_width * ss_factor

        start_time = self.end_time - self.LINE_DUR * height
        sat_positions = np.empty((height, 2), dtype=float)
        self._x_offsets = np.empty(height, dtype=float)
        x_fovs = np.empty(height, dtype=float)

        a = math.radians(self.NOAA_AVHRR_FOV / 2)
        b = math.pi / 2 - a
        sa = math.sin(a)
        for i in range(height):
            t = start_time + self.LINE_DUR * i
            self.sat_ephem.compute(t)
            sat_positions[i] = self.sat_ephem.sublong, self.sat_ephem.sublat

            c = b - math.acos(sa * (ephem.earth_radius + self.sat_ephem.elevation) / ephem.earth_radius)
            x_fovs[i] = c * 2 / 909

        self._ref_lonlat = sat_positions[height // 2]
        self._y_res = (ephem.separation(sat_positions[0], sat_positions[-1]) / height) / ss_factor
        # x_res = np.mean(x_fovs)
        self._ref_x_res = x_fovs[height // 2] / ss_factor
        self._ref_az = utils.azimuth(self._ref_lonlat, sat_positions[height // 2 + 1])

        for i in range(height):
            self._x_offsets[i] = self._lonlat_to_rel_px(x_fovs[i], sat_positions[i])[0]

        overlay_img = Image.new('RGBA', (self._ss_width, self._ss_height))
        self._draw = ImageDraw.Draw(overlay_img)

        for points, color in shapes.iter():
            if isinstance(points, dict):
                if points['name'] == 'observer':
                    points['lonlat'] = np.radians(self.observer_lonlat)
                self._draw_fig(points, ss_factor)
            else:
                self._draw_lines(points, color)

        self.map_overlay = np.array(overlay_img.resize((self.IMAGE_WIDTH, height), Image.Resampling.LANCZOS), dtype=np.uint8)

        del self._height
        del self._ss_height
        del self._half_ss_height
        del self._ss_width
        del self._half_ss_width
        del self._line_width
        del self._x_offsets
        del self._ref_lonlat
        del self._y_res
        del self._ref_x_res
        del self._ref_az
        del self._draw

    def _lonlat_to_rel_px(self, x_res, lonlat):
        ab = utils.azimuth(self._ref_lonlat, lonlat) - self._ref_az
        c = min(max(ephem.separation(self._ref_lonlat, lonlat), -utils.THIRD_PI), utils.THIRD_PI)
        a = math.atan(math.cos(ab) * math.tan(c))
        b = math.asin(math.sin(ab) * math.sin(c))
        x = -b / x_res
        y = a / self._y_res
        return x, y

    def _draw_lines(self, points, color):
        to_draw = [(0, 0)] * len(points)

        for i, pt in enumerate(points):
            x, y = self._lonlat_to_rel_px(self._ref_x_res, pt)
            x -= self._x_offsets[round(np.clip(y, 0, self._height - 1))]
            to_draw[i] = x + self._half_ss_width, y + self._half_ss_height

        self._draw.line(to_draw, fill=color, width=self._line_width)

    def _draw_fig(self, fig: dict, ss_factor):
        x, y = self._lonlat_to_rel_px(self._ref_x_res, fig['lonlat'])
        x -= self._x_offsets[round(np.clip(y, 0, self._height - 1))]
        x += self._half_ss_width
        y += self._half_ss_height

        col = fig['color']
        sz = np.multiply(fig['size'], ss_factor)
        if fig['type'] == '+':
            lwidth, llen = sz
            llen /= 2
            self._draw.line(((x - llen, y), (x + llen, y)), fill=col, width=lwidth)
            self._draw.line(((x, y - llen), (x, y + llen)), fill=col, width=lwidth)

        elif fig['type'] == 'o':
            self._draw.ellipse(((x - sz, y - sz), (x + sz, y + sz)), fill=col)

    def black_overlay(self):
        return self.map_overlay * np.array([0, 0, 0, 1], dtype=np.uint8)



import signal

from unittest import TestCase

from sats_receiver.async_signal import AsyncSignal


class TestAsyncSignal(TestCase):
    def test_wait_sigint(self):
        signame = 'SIGINT'
        with AsyncSignal() as asig:
            signal.raise_signal(signal.SIGINT)

            self.assertEqual(signame, asig.wait(1))

    def test_sig_invalid(self):
        signame = 'SIGAAA'

        with AsyncSignal([signame]) as asig:
            with self.assertRaises(AttributeError):
                asig.wait(1)



import datetime as dt
import logging
import math
import pathlib

from typing import Optional, Union

import dateutil.tz
import ephem
import gnuradio as gr
import gnuradio.analog
import gnuradio.blocks
import gnuradio.digital
import gnuradio.fft
import gnuradio.filter
import gnuradio.gr

from PIL import ExifTags

from sats_receiver import utils
from sats_receiver.gr_modules.epb import sstv as sstv_epb
from sats_receiver.observer import Observer
from sats_receiver.systems import apt, sstv


class Decoder(gr.gr.hier_block2):
    def __init__(self,
                 name: str,
                 sat_name: str,
                 samp_rate: Union[int, float],
                 out_dir: pathlib.Path):
        self.prefix = f'{self.__class__.__name__}: {sat_name}'
        self.log = logging.getLogger(self.prefix)

        super(Decoder, self).__init__(
            name,
            gr.gr.io_signature(1, 1, gr.gr.sizeof_gr_complex),
            gr.gr.io_signature(0, 0, 0)
        )

        self.now = dt.datetime.fromtimestamp(0)
        self.sat_name = sat_name
        self.samp_rate = samp_rate
        self.out_dir = out_dir
        self.tmp_file = utils.mktmp(prefix='_'.join(name.lower().split()))
        self.base_kw = dict(log=self.log, sat_name=sat_name, out_dir=out_dir)

    def start(self):
        self.tmp_file = self.out_dir / ('_'.join([self.t.strftime('%Y%m%d%H%M%S'),
                                                  *self.name().lower().split()]) + '.tmp')
        self.base_kw.update(tmp_file=self.tmp_file)

    def finalize(self, executor, fin_key: str):
        pass

    @property
    def t(self) -> dt.datetime:
        t = dt.datetime.now()
        self.now = t
        return t


class RawDecoder(Decoder):
    def __init__(self,
                 sat_name: str,
                 samp_rate: Union[int, float],
                 out_dir: pathlib.Path):
        super(RawDecoder, self).__init__('Raw Decoder', sat_name, samp_rate, out_dir)

        self.ctf = gr.blocks.complex_to_float(1)
        self.wav_sink = gr.blocks.wavfile_sink(
            str(self.tmp_file),
            2,
            samp_rate,
            gr.blocks.FORMAT_WAV,
            gr.blocks.FORMAT_FLOAT,
            False
        )
        self.wav_sink.close()
        utils.unlink(self.tmp_file)

        self.connect(self, self.ctf, self.wav_sink)
        self.connect((self.ctf, 1), (self.wav_sink, 1))

    def start(self):
        super(RawDecoder, self).start()

        self.wav_sink.open(str(self.tmp_file))

    def finalize(self, executor, fin_key: str):
        self.wav_sink.close()
        if self.tmp_file.exists():
            executor.execute(self._raw_finalize, **self.base_kw, fin_key=fin_key)

    @staticmethod
    def _raw_finalize(log, sat_name, out_dir, tmp_file, fin_key) -> Optional[tuple[str, str, str, dt.datetime]]:
        log.debug('finalizing...')

        d = dt.datetime.fromtimestamp(tmp_file.stat().st_mtime, dateutil.tz.tzutc())
        res_fn = tmp_file.rename(out_dir / d.strftime(f'{sat_name}_%Y-%m-%d_%H-%M-%S_RAW.wav'))
        st = res_fn.stat()
        log.info('finish: %s (%s)', res_fn, utils.numbi_disp(st.st_size))

        return sat_name, fin_key, res_fn, dt.datetime.fromtimestamp(st.st_mtime, dateutil.tz.tzutc())


class AptDecoder(Decoder):
    def __init__(self,
                 sat_name: str,
                 samp_rate: Union[int, float],
                 out_dir: pathlib.Path,
                 sat_ephem_tle: Optional[tuple[ephem.EarthSatellite, tuple[str, str, str]]],
                 observer_lonlat: tuple[float, float]):
        name = 'APT Decoder'
        super(AptDecoder, self).__init__(name, sat_name, samp_rate, out_dir)

        pfx = '_'.join(name.lower().split())
        self.corr_file = utils.mktmp(dir=out_dir, prefix=pfx, suffix='.corr')
        self.peaks_file = utils.mktmp(dir=out_dir, prefix=pfx, suffix='.peaks')
        self.sat_ephem_tle = sat_ephem_tle
        self.observer_lonlat = observer_lonlat
        self.base_kw.update(sat_tle=sat_ephem_tle[1], observer_lonlat=observer_lonlat)

        resamp_gcd = math.gcd(samp_rate, apt.Apt.WORK_RATE)

        self.frs = gr.blocks.rotator_cc(2 * math.pi * -apt.Apt.CARRIER_FREQ / samp_rate)
        self.rsp = gr.filter.rational_resampler_ccc(
            interpolation=apt.Apt.WORK_RATE // resamp_gcd,
            decimation=samp_rate // resamp_gcd,
            taps=[],
            fractional_bw=0,
        )

        tofilter_freq = 2080
        lp_gcd = math.gcd(apt.Apt.WORK_RATE, tofilter_freq)

        self.lpf = gr.filter.fir_filter_ccf(
            1,
            gr.filter.firdes.low_pass(
                gain=1,
                sampling_freq=apt.Apt.WORK_RATE / lp_gcd,
                cutoff_freq=tofilter_freq / lp_gcd,
                transition_width=0.1,
                window=gr.fft.window.WIN_HAMMING,
                param=6.76,
            )
        )
        self.ctm = gr.blocks.complex_to_mag()
        self.ftc = gr.blocks.float_to_complex()
        # self.dc_rem = gr.blocks.correctiq()
        self.correllator = gr.digital.corr_est_cc(
            symbols=apt.Apt.SYNC_A,
            sps=1,
            mark_delay=1,
            threshold=0.9,
            threshold_method=gr.digital.THRESHOLD_ABSOLUTE,
        )
        self.ctf_out = gr.blocks.complex_to_float()
        self.ctf_corr = gr.blocks.complex_to_float()

        self.out_file_sink = gr.blocks.file_sink(gr.gr.sizeof_float, str(self.tmp_file), False)
        self.out_corr_sink = gr.blocks.file_sink(gr.gr.sizeof_float, str(self.corr_file), False)
        self.peak_detector = gr.blocks.peak_detector2_fb(
            threshold_factor_rise=7,
            look_ahead=apt.Apt.FRAME_WIDTH * apt.Apt.WORK_RATE // apt.Apt.FINAL_RATE,
            alpha=0.001,
        )
        self.out_peaks_sink = gr.blocks.file_sink(gr.gr.sizeof_char, str(self.peaks_file), False)

        self.connect(
            self,
            self.frs,
            self.rsp,
            self.lpf,
            self.ctm,
            self.ftc,
            # self.dc_rem,
            self.correllator,
            self.ctf_out,
            self.out_file_sink,
        )
        self.connect(
            (self.correllator, 1),
            self.ctf_corr,
            self.out_corr_sink,
        )
        self.connect(
            self.ctf_corr,
            self.peak_detector,
            self.out_peaks_sink,
        )

        utils.close(self.out_file_sink, self.out_corr_sink, self.out_peaks_sink)
        utils.unlink(self.tmp_file, self.corr_file, self.peaks_file)

    def start(self):
        super(AptDecoder, self).start()
        self.corr_file = self.out_dir / ('_'.join([self.now.strftime('%Y%m%d%H%M%S'),
                                                   *self.name().lower().split()]) + '.corr.tmp')
        self.peaks_file = self.out_dir / ('_'.join([self.now.strftime('%Y%m%d%H%M%S'),
                                                    *self.name().lower().split()]) + '.peaks.tmp')
        self.base_kw.update(corr_file=self.corr_file, peaks_file=self.peaks_file)

        self.out_file_sink.open(str(self.tmp_file))
        self.out_file_sink.set_unbuffered(False)

        self.out_corr_sink.open(str(self.corr_file))
        self.out_corr_sink.set_unbuffered(False)

        self.out_peaks_sink.open(str(self.peaks_file))
        self.out_peaks_sink.set_unbuffered(False)

    def finalize(self, executor, fin_key: str):
        self.out_file_sink.do_update()
        self.out_corr_sink.do_update()
        self.out_peaks_sink.do_update()

        utils.close(self.out_file_sink, self.out_corr_sink, self.out_peaks_sink)

        for p in self.tmp_file, self.corr_file, self.peaks_file:
            if not p.exists():
                self.log.warning('%s: missing components `%s`', p)
                utils.unlink(self.tmp_file, self.corr_file, self.peaks_file)
                return

        executor.execute(self._apt_finalize, **self.base_kw, fin_key=fin_key)

    @staticmethod
    def _apt_finalize(log: logging.Logger,
                      sat_name: str,
                      sat_tle: tuple[str, str, str],
                      observer_lonlat: tuple[float, float],
                      tmp_file: pathlib.Path,
                      corr_file: pathlib.Path,
                      peaks_file: pathlib.Path,
                      out_dir: pathlib.Path,
                      fin_key: str) -> Optional[tuple[str, str, pathlib.Path, dt.datetime]]:
        log.debug('finalizing...')

        a = apt.Apt(sat_name, tmp_file, corr_file, peaks_file, sat_tle, observer_lonlat)
        x = a.process()

        utils.unlink(tmp_file, corr_file, peaks_file)

        if x:
            log.info('finish with error')
        else:
            res_fn, sz = a.to_apt(out_dir)
            log.info('finish: %s (%s)', res_fn, utils.numbi_disp(sz))

            return sat_name, fin_key, res_fn, a.end_time


class RawStreamDecoder(Decoder):
    def __init__(self,
                 sat_name: str,
                 samp_rate: Union[int, float],
                 out_dir: pathlib.Path,
                 name='RAW Stream Decoder'):
        super(RawStreamDecoder, self).__init__(name, sat_name, samp_rate, out_dir)

        self.ctf = gr.blocks.complex_to_float(1)
        self.rail = gr.analog.rail_ff(-1, 1)
        self.ftch = gr.blocks.float_to_char(1, 127)
        # self.fts = gr.blocks.float_to_short(1, 32767)
        # self.stch = gr.blocks.short_to_char(1)

        self.out_file_sink = gr.blocks.file_sink(gr.gr.sizeof_char, str(self.tmp_file), False)
        self.out_file_sink.close()
        utils.unlink(self.tmp_file)

        self.connect(
            self,
            self.ctf,
            self.rail,
            self.ftch,
            self.out_file_sink,
        )

    def start(self):
        super(RawStreamDecoder, self).start()

        self.out_file_sink.open(str(self.tmp_file))
        self.out_file_sink.set_unbuffered(False)

    def finalize(self, executor, fin_key: str):
        self.out_file_sink.do_update()
        self.out_file_sink.close()
        if self.tmp_file.exists():
            executor.execute(self._raw_stream_finalize, **self.base_kw, fin_key=fin_key)

    @staticmethod
    def _raw_stream_finalize(log: logging.Logger,
                             sat_name: str,
                             out_dir: pathlib.Path,
                             tmp_file: pathlib.Path,
                             fin_key: str) -> Optional[tuple[str, str, pathlib.Path, dt.datetime]]:
        log.debug('finalizing...')

        d = dt.datetime.fromtimestamp(tmp_file.stat().st_mtime, dateutil.tz.tzutc())
        res_fn = tmp_file.rename(out_dir / d.strftime(f'{sat_name}_%Y-%m-%d_%H-%M-%S_RAW.s'))
        st = res_fn.stat()
        log.info('finish: %s (%s)', res_fn, utils.numbi_disp(st.st_size))

        return sat_name, fin_key, res_fn, dt.datetime.fromtimestamp(st.st_mtime, dateutil.tz.tzutc())


class LrptDecoder(RawStreamDecoder):
    def __init__(self,
                 sat_name: str,
                 samp_rate: Union[int, float],
                 out_dir: pathlib.Path):
        raise NotImplementedError()
        super(LrptDecoder, self).__init__(sat_name, samp_rate, out_dir, name='LRPT Decoder')

    def start(self):
        super(LrptDecoder, self).start()

    def finalize(self, executor, fin_key: str):
        super(LrptDecoder, self).finalize(executor, fin_key)


class SstvDecoder(Decoder):
    _FREQ_1 = (1900 - 1200) / 2

    def __init__(self,
                 sat_name: str,
                 samp_rate: Union[int, float],
                 out_dir: pathlib.Path,
                 observer: Observer = None,
                 do_sync=True,
                 wsr=16000):
        super(SstvDecoder, self).__init__('SSTV Decoder', sat_name, samp_rate, out_dir)

        self.observer = observer
        self.do_sync = do_sync
        self.wsr = wsr

        hdr_pix_width = int(wsr * sstv.Sstv.HDR_PIX_S)
        hdr = sstv.Sstv.HDR_SYNC_WORD.repeat(hdr_pix_width)
        resamp_gcd = math.gcd(wsr, samp_rate)

        self.bpf_input = gr.filter.fir_filter_ccc(
            1,
            gr.filter.firdes.complex_band_pass(
                1,          # gain
                samp_rate,  # sampling_freq
                900,        # low_cutoff_freq
                2500,       # high_cutoff_freq
                200,        # transition_width
                gr.fft.window.WIN_HAMMING,
                6.76))
        self.frs = gr.blocks.rotator_cc(2 * math.pi * -(1900 - self._FREQ_1) / samp_rate)
        self.quad_demod = gr.analog.quadrature_demod_cf((samp_rate / (2 * math.pi * self._FREQ_1)))
        self.rsp = gr.filter.rational_resampler_fcc(
                interpolation=wsr // resamp_gcd,
                decimation=samp_rate // resamp_gcd,
                taps=[],
                fractional_bw=0
        )
        self.correllator = gr.digital.corr_est_cc(hdr, hdr_pix_width, 1, 0.4, gr.digital.THRESHOLD_ABSOLUTE)
        self.ctf_out = gr.blocks.complex_to_float()
        self.out_add_const = gr.blocks.add_const_ff(450 / self._FREQ_1)
        self.out_multiply_const = gr.blocks.multiply_const_ff(self._FREQ_1 / (750 + 450))
        self.ctf_corr = gr.blocks.complex_to_float()
        self.corr_peak_detector = gr.blocks.peak_detector2_fb(0.1, hdr.size, 0.001)
        self.sstv_epb = sstv_epb.SstvEpb(wsr, do_sync, self.log, sat_name, out_dir)

        self.connect(
            self,
            self.bpf_input,
            self.frs,
            self.quad_demod,
            self.rsp,
            self.correllator,
            self.ctf_out,
            self.out_add_const,
            self.out_multiply_const,
            (self.sstv_epb, self.sstv_epb.OUT_IN),
        )
        self.connect(
            (self.correllator, 1),
            self.ctf_corr,
            self.corr_peak_detector,
            (self.sstv_epb, self.sstv_epb.PEAKS_IN),
        )

        if observer is None:
            latlonalt = None
        else:
            latlonalt = str(observer.lat), str(observer.lon), observer.elev
        self.base_kw.update(observer_latlonalt=latlonalt)

    def start(self):
        # super(SstvDecoder, self).start()
        utils.unlink(self.tmp_file)

    def finalize(self, executor, fin_key: str):
        self.sstv_epb.stop()
        sstv_rr: list[sstv.SstvRecognizer] = self.sstv_epb.finalize()
        executor.execute(self._sstv_finalize, **self.base_kw, sstv_rr=sstv_rr, fin_key=fin_key)

    @staticmethod
    def _sstv_finalize(log: logging.Logger,
                       sat_name: str,
                       out_dir: pathlib.Path,
                       observer_latlonalt: Optional[tuple[str, str, Union[int, float]]],
                       sstv_rr: list[sstv.SstvRecognizer],
                       fin_key: str) -> tuple[str, str, list[tuple[pathlib.Path, dt.datetime]]]:
        log.debug('finalizing...')

        if observer_latlonalt:
            observer = ephem.Observer()
            observer.lat, observer.lon, observer.elev = observer_latlonalt

        fn_dt = []
        sz_sum = 0
        for i in sstv_rr:
            img = i.get_image()
            if not img:
                continue

            if observer_latlonalt:
                log.debug('add GPSInfo EXIF')
                img = utils.img_add_exif(img, observer=observer)
            exif = img.getexif()

            sstv_mode = exif.get_ifd(ExifTags.IFD.Exif).get(ExifTags.Base.UserComment, 'SSTV')
            end_time = exif.get(ExifTags.Base.DateTime)
            end_time = (dt.datetime.strptime(end_time, '%Y:%m:%d %H:%M:%S')
                        if end_time
                        else dt.datetime.utcnow())
            res_fn = out_dir / end_time.strftime(f'{sat_name}_{sstv_mode}_%Y-%m-%d_%H-%M-%S,%f.png')

            img.save(res_fn, exif=exif)

            fn_dt.append((res_fn, end_time))
            sz = res_fn.stat().st_size
            sz_sum += sz
            log.info('finish: %s (%s)', res_fn, utils.numbi_disp(sz))

        log.info('finish %s images (%s)', len(fn_dt), utils.numbi_disp(sz_sum))

        return sat_name, fin_key, fn_dt



import atexit
import datetime as dt
import logging
import logging.handlers
import multiprocessing as mp
import multiprocessing.connection
import pathlib
import tempfile
import time

from hashlib import sha256
from test import support
from unittest import TestCase

import ephem
import gnuradio as gr
import gnuradio.blocks
import gnuradio.gr
import numpy as np

from PIL import Image, ExifTags
from sats_receiver import utils
from sats_receiver.manager import Executor
from sats_receiver.gr_modules.decoders import Decoder, SstvDecoder
from sats_receiver.gr_modules.epb.prober import Prober
from sats_receiver.observer import Observer


HERE = pathlib.Path(__file__).parent
FILES = HERE / 'files'
TIMEOUT = support.SHORT_TIMEOUT


class DecoderExecutor(Executor):
    def __init__(self, ret_wr: mp.connection.Connection):
        super().__init__()
        self.ret_wr = ret_wr

    def action(self):
        while 1:
            self.sysu.collect()

            try:
                x = self.rd.poll(1)
            except InterruptedError:
                x = 1
            except:
                continue

            if not x:
                continue

            x = self.rd.recv()

            if x == '.':
                break

            try:
                fn, args, kwargs = x
            except ValueError:
                self.log.error('invalid task: %s', x)
                continue

            if callable(fn):
                try:
                    x = fn(*args, **kwargs)
                except Exception:
                    self.log.exception('%s with args=%s kwargs=%s', fn, args, kwargs)
                    continue

                if x and isinstance(x, tuple):
                    if len(x) == 4:
                        sat_name, fin_key, res_filename, end_time = x
                    elif len(x) == 3:
                        sat_name, fin_key, fn_dt = x

                self.ret_wr.send(x)


class DecoderTopBlock(gr.gr.top_block):
    def __init__(self,
                 wav_fp: pathlib.Path,
                 ret_wr: mp.connection.Connection,
                 decoder: Decoder):
        self.prefix = self.__class__.__name__
        self.log = logging.getLogger(self.prefix)

        super(DecoderTopBlock, self).__init__('DecoderTopBlock', catch_exceptions=False)

        self.executor = DecoderExecutor(ret_wr)

        self.blocks_wavfile_source = gr.blocks.wavfile_source(str(wav_fp), False)
        self.prober = Prober()
        self.blocks_float_to_complex = gr.blocks.float_to_complex(1)
        self.decoder = decoder

        self.connect(
            self.blocks_wavfile_source,
            self.blocks_float_to_complex,
            self.decoder,
        )
        self.connect(
            self.blocks_wavfile_source,
            self.prober,
        )

    def start(self, max_noutput_items=10000000):
        self.log.info('START')
        self.executor.start()
        atexit.register(lambda x: (x.stop(), x.join()), self.executor)
        self.decoder.start()
        super(DecoderTopBlock, self).start(max_noutput_items)

    def stop(self):
        self.log.info('STOP')
        super(DecoderTopBlock, self).stop()

        fin_key = sha256((self.prefix + str(dt.datetime.now())).encode()).hexdigest()
        self.decoder.finalize(self.executor, fin_key)

        self.executor.stop()

    def wait(self):
        super(DecoderTopBlock, self).wait()
        self.executor.join()


class TestDecoders(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.out_dir = tempfile.TemporaryDirectory('.d', 'sats-receiver-test', ignore_cleanup_errors=True)
        cls.out_dp = pathlib.Path(cls.out_dir.name)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.out_dir.cleanup()
        cls.out_dir = None

    def setUp(self) -> None:
        self.ret_rd, self.ret_wr = mp.Pipe(False)
        self.tb = None
        self.to_close = [self.ret_rd, self.ret_wr]

    def tearDown(self) -> None:
        if isinstance(self.tb, DecoderTopBlock):
            self.tb.stop()
            self.tb.wait()
        utils.close(*self.to_close)

    def test_sstv_robot36(self):
        lat = 11.111
        lon = -22.222
        alt = -33.333
        sstv_wsr = 16000
        wav_fp = FILES / 'Robot36_16kHz.wav'
        wav_samp_rate = 16000

        decoder = SstvDecoder(
            sat_name='Test Sat',
            samp_rate=wav_samp_rate,
            out_dir=self.out_dp,
            observer=Observer({'latitude': lat,
                               'longitude': lon,
                               'elevation': alt,
                               'weather': False}),
            do_sync=1,
            wsr=sstv_wsr,
        )
        self.tb = DecoderTopBlock(wav_fp, self.ret_wr, decoder)
        self.tb.start()

        while self.tb.prober.changes():
            time.sleep(self.tb.prober.measure_s)

        self.tb.stop()
        self.tb.wait()

        x = self.ret_rd.poll(TIMEOUT)
        self.assertTrue(x)

        x = self.ret_rd.recv()
        self.assertIsInstance(x, tuple)
        self.assertEqual(len(x), 3)

        sat_name, fin_key, fn_dts = x
        self.assertEqual(sat_name, 'Test Sat')
        self.assertIsInstance(fn_dts, list)
        self.assertEqual(len(fn_dts), 1)

        fn_dt = fn_dts[0]
        self.assertIsInstance(fn_dt, tuple)
        self.assertEqual(len(fn_dt), 2)

        fp, d = fn_dt
        self.assertIsInstance(fp, pathlib.Path)
        self.assertTrue(fp.is_file())
        self.assertIsInstance(d, dt.datetime)

        img = Image.open(fp)
        self.to_close.append(img)
        exif = img.getexif()
        ee = exif.get_ifd(ExifTags.IFD.Exif)
        gps = exif.get_ifd(ExifTags.IFD.GPSInfo)

        self.assertEqual(ee.get(ExifTags.Base.UserComment), 'Robot36')
        end_time = dt.datetime.strptime(exif.get(ExifTags.Base.DateTime), '%Y:%m:%d %H:%M:%S')
        self.assertEqual(end_time, d)

        self.assertEqual(ephem.degrees('{}:{}:{}'.format(*gps[ExifTags.GPS.GPSLatitude])),
                         ephem.degrees(str(abs(lat))))
        self.assertEqual(gps[ExifTags.GPS.GPSLatitudeRef], 'N')

        self.assertEqual(ephem.degrees('{}:{}:{}'.format(*gps[ExifTags.GPS.GPSLongitude])),
                         ephem.degrees(str(abs(lon))))
        self.assertEqual(gps[ExifTags.GPS.GPSLongitudeRef], 'W')

        self.assertTrue(np.isclose(float(gps[ExifTags.GPS.GPSAltitude]), abs(alt)))
        self.assertEqual(gps[ExifTags.GPS.GPSAltitudeRef], b'\1')



import sys

from ctypes import CDLL, c_char, c_char_p, c_int, c_uint32, c_void_p, POINTER


if sys.platform == 'win32':
    _lib = CDLL('rtlsdr')
else:
    _lib = CDLL('librtlsdr.so.0')


class LibRtlSdrError(RuntimeError):
    pass


class LibUsbError(RuntimeError):
    CODES = {
        0: 'Success (no error)',
        -1: 'Input/output error',
        -2: 'Invalid parameter',
        -3: 'Access denied (insufficient permissions)',
        -4: 'No such device (it may have been disconnected)',
        -5: 'Entity not found',
        -6: 'Resource busy',
        -7: 'Operation timed out',
        -8: 'Overflow',
        -9: 'Pipe error',
        -10: 'System call interrupted (perhaps due to signal)',
        -11: 'Insufficient memory',
        -12: 'Operation not supported or unimplemented on this platform',
        -99: 'Other error',
    }

    def __init__(self, err_code):
        super(LibUsbError, self).__init__(f'[{err_code}] {self.CODES.get(err_code, -99)}')


rtlsdr_dev_p = c_void_p

_f = _lib.rtlsdr_open
_f.argtypes = POINTER(rtlsdr_dev_p), c_uint32
_f.restype = c_int
def rtlsdr_open(idx=0) -> rtlsdr_dev_p:
    dev = rtlsdr_dev_p(None)
    if _lib.rtlsdr_open(dev, idx):
        raise LibRtlSdrError('Could not open RTL-SDR')
    return dev


_f = _lib.rtlsdr_set_bias_tee_gpio
_f.argtypes = rtlsdr_dev_p, c_int, c_int
_f.restype = c_int
def rtlsdr_set_bias_tee_gpio(dev: rtlsdr_dev_p, pin: int, on: int):
    if _lib.rtlsdr_set_bias_tee_gpio(dev, pin, on):
        raise LibRtlSdrError('Invalid device')


_f = _lib.rtlsdr_close
_f.argtypes = rtlsdr_dev_p,
_f.restype = c_int
def rtlsdr_close(dev: rtlsdr_dev_p):
    _lib.rtlsdr_close(dev)


_f = _lib.rtlsdr_get_device_count
_f.restype = c_uint32
def rtlsdr_get_device_count() -> int:
    return _lib.rtlsdr_get_device_count()

_f = _lib.rtlsdr_get_device_name
_f.argtypes = c_uint32,
_f.restype = c_char_p
def rtlsdr_get_device_name(idx: int) -> str:
    return _lib.rtlsdr_get_device_name(idx).decode('utf8')


_f = _lib.rtlsdr_get_device_usb_strings
_f.argtypes = c_uint32, POINTER(c_char), POINTER(c_char), POINTER(c_char)
_f.restype = c_int
def rtlsdr_get_device_usb_strings(index: int) -> tuple[str, str, str]:
    manufact = (c_char * 256)()
    product = (c_char * 256)()
    serial = (c_char * 256)()

    r = _lib.rtlsdr_get_device_usb_strings(index, manufact, product, serial)
    if r:
        raise LibUsbError(r)

    return manufact.value.decode('utf8'), product.value.decode('utf8'), serial.value.decode('utf8')


_f = _lib.rtlsdr_get_index_by_serial
_f.argtypes = POINTER(c_char),
_f.restype = c_int
def rtlsdr_get_index_by_serial(serial: str) -> int:
    r = _lib.rtlsdr_get_index_by_serial(serial.encode('utf8'))
    if r >= 0:
        return r

    # values below gettings from library source code
    if r == -1:
        m = 'Null serial'
    elif r == -2:
        m = 'No devices'
    elif r == -3:
        m = 'Device not found'
    else:   # unreachable
        m = f'Unknown {r}'

    raise LibRtlSdrError(m)


def get_serials() -> str:
    for i in range(rtlsdr_get_device_count()):
        yield rtlsdr_get_device_usb_strings(i)[2]


def set_bt(bt: int, serial: str = None, pin=0):
    """
    Set Bias-T state

    :param bt: Bias-t value (1 - enable, 0 - disable)
    :param serial: Serial Number of the device that needs to enable bias-t
    :param pin: GPIO pin to change bias-t state. 0 by default
    """
    i = serial and rtlsdr_get_index_by_serial(serial) or 0
    dev = rtlsdr_open(i)
    rtlsdr_set_bias_tee_gpio(dev, pin, bt)
    rtlsdr_close(dev)



import datetime as dt
import logging
import math
import pathlib

from hashlib import sha256
from typing import Mapping, Optional, Union

import ephem
import gnuradio as gr
import gnuradio.analog
import gnuradio.blocks
import gnuradio.filter
import gnuradio.gr

from sats_receiver import utils
from sats_receiver.gr_modules import decoders, demodulators
from sats_receiver.observer import Observer


class RadioModule(gr.gr.hier_block2):
    def __init__(self,
                 main_tune: Union[int, float],
                 samp_rate: Union[int, float],
                 bandwidth: Union[int, float],
                 frequency: Union[int, float]):
        super(RadioModule, self).__init__(
            'Radio Module',
            gr.gr.io_signature(1, 1, gr.gr.sizeof_gr_complex),
            gr.gr.io_signature(1, 1, gr.gr.sizeof_gr_complex)
        )

        self.enabled = False
        self.main_tune = main_tune
        self.samp_rate = samp_rate
        self.bandwidth = bandwidth
        self.frequency = frequency

        self.resamp_gcd = resamp_gcd = math.gcd(bandwidth, samp_rate)

        self.blocks_copy = gr.blocks.copy(gr.gr.sizeof_gr_complex)
        self.blocks_copy.set_enabled(self.enabled)
        self.freqshifter = gr.blocks.rotator_cc(2 * math.pi * -(main_tune - frequency) / samp_rate)
        self.resampler = gr.filter.rational_resampler_ccc(
            interpolation=bandwidth // resamp_gcd,
            decimation=samp_rate // resamp_gcd,
            taps=[],
            fractional_bw=0,
        )

        self.connect(
            self,
            self.blocks_copy,
            self.freqshifter,
            self.resampler,
            self,
        )

    def set_enabled(self, enabled):
        self.enabled = enabled
        self.blocks_copy.set_enabled(enabled)

    def set_freq_offset(self, new_freq: Union[int, float]):
        self.freqshifter.set_phase_inc(2 * math.pi * (self.main_tune - new_freq) / self.samp_rate)


class SatRecorder(gr.gr.hier_block2):
    @staticmethod
    def _validate_config(config: Mapping) -> bool:
        return (
            all(map(lambda x: x in config,
                    [
                        # 'enabled',          # optional
                        'freq',
                        # 'freq_correction',  # optional
                        'bandwidth',
                        'mode',
                        # 'decode',           # optional

                        # 'qpsk_baudrate',    # only in QPSK demode
                        # 'qpsk_excess_bw',   # optional
                        # 'qpsk_ntaps',       # optional
                        # 'qpsk_costas_bw',   # optional

                        # 'sstv_wsr',         # optional, only in SSTV decode
                        # 'sstv_sync',        # optional, only in SSTV decode
                    ]))
            and (config['mode'] != utils.Mode.QPSK or 'qpsk_baudrate' in config)
        )

    def __init__(self,
                 up: 'Satellite',
                 config: Mapping,
                 main_tune: Union[int, float],
                 samp_rate: Union[int, float]):
        f = config.get('freq')
        self.prefix = f'{self.__class__.__name__}: {up.name}: {f and f": {utils.num_disp(f)}Hz"}'

        if not self._validate_config(config):
            raise ValueError(f'{self.prefix}: Invalid config!')

        super(SatRecorder, self).__init__(
            self.prefix,
            gr.gr.io_signature(1, 1, gr.gr.sizeof_gr_complex),
            gr.gr.io_signature(0, 0, 0)
        )

        self.config = config
        self.radio = RadioModule(main_tune, samp_rate, self.bandwidth, self.frequency)
        self.demodulator = None
        self.post_demod = gr.blocks.float_to_complex()

        try:
            self.mode == self.decode
        except ValueError as e:
            if 'Decode' in str(e):
                x = 'decoder', self.decode
            else:
                x = 'demodulation', self.mode
            raise ValueError(f'{self.prefix}: Unknown {x[0]} `{x[1]}` for `{up.name}`')

        if self.mode == utils.Mode.AM:
            self.demodulator = gr.analog.am_demod_cf(
                channel_rate=self.bandwidth,
                audio_decim=1,
                audio_pass=5000,
                audio_stop=5500,
            )

        elif self.mode == utils.Mode.FM:
            self.demodulator = gr.analog.fm_demod_cf(
                channel_rate=self.bandwidth,
                audio_decim=1,
                deviation=self.bandwidth,
                audio_pass=16000,
                audio_stop=17000,
                gain=1.0,
                tau=0,
            )

        elif self.mode == utils.Mode.WFM:
            self.demodulator = gr.analog.wfm_rcv(
                quad_rate=self.bandwidth,
                audio_decimation=1,
            )

        elif self.mode == utils.Mode.WFM_STEREO:
            if self.bandwidth < 76800:
                raise ValueError(f'{self.prefix}: param `bandwidth` for WFM Stereo must be at least 76800, got {self.bandwidth} instead')
            self.demodulator = gr.analog.wfm_rcv_pll(
                demod_rate=self.bandwidth,
                audio_decimation=1,
                deemph_tau=50e-6,
            )
            self.connect((self.demodulator, 1), (self.post_demod, 1))

        elif self.mode == utils.Mode.QUAD:
            self.demodulator = gr.analog.quadrature_demod_cf(1)

        elif self.mode == utils.Mode.QPSK:
            self.demodulator = demodulators.QpskDemod(self.bandwidth, self.qpsk_baudrate, self.qpsk_excess_bw, self.qpsk_ntaps, self.qpsk_costas_bw)

        if self.decode == utils.Decode.APT:
            self.decoder = decoders.AptDecoder(up.name, self.bandwidth, up.output_directory, up.sat_ephem_tle, up.observer.lonlat)

        # elif self.decode == Decode.LRPT:
        #     # TODO
        #     # self.decoder =

        elif self.decode == utils.Decode.RSTREAM:
            self.decoder = decoders.RawStreamDecoder(up.name, self.bandwidth, up.output_directory)

        elif self.decode == utils.Decode.RAW:
            self.decoder = decoders.RawDecoder(up.name, self.bandwidth, up.output_directory)

        elif self.decode == utils.Decode.SSTV:
            self.decoder = decoders.SstvDecoder(up.name, self.bandwidth, up.output_directory, up.observer, self.sstv_sync, self.sstv_wsr)

        self.connect(
            self,
            self.radio,
            *((self.demodulator, self.post_demod) if self.demodulator else tuple()),
            self.decoder,
        )

    def set_freq_offset(self, new_freq: Union[int, float]):
        self.radio.set_freq_offset(new_freq)

    @property
    def is_runned(self) -> bool:
        return self.radio.enabled

    @property
    def enabled(self) -> bool:
        return self.config.get('enabled', True)

    @property
    def freq(self) -> Union[int, float]:
        return self.config['freq']

    @property
    def freq_correction(self) -> Union[int, float]:
        return self.config.get('freq_correction', 0)

    @property
    def frequency(self) -> Union[int, float]:
        return self.freq + self.freq_correction

    @property
    def bandwidth(self) -> Union[int, float]:
        return self.config['bandwidth']

    @property
    def mode(self) -> utils.Mode:
        return utils.Mode(self.config.get('mode', utils.Mode.RAW.value))

    @property
    def decode(self) -> utils.Decode:
        return utils.Decode(self.config.get('decode', utils.Decode.RAW.value))

    @property
    def qpsk_baudrate(self) -> Union[int, float]:
        return self.config['qpsk_baudrate']

    @property
    def qpsk_excess_bw(self) -> Union[int, float]:
        return self.config.get('qpsk_excess_bw', 0.35)

    @property
    def qpsk_ntaps(self) -> int:
        return int(self.config.get('qpsk_ntaps'), 33)

    @property
    def qpsk_costas_bw(self) -> Union[int, float]:
        return self.config.get('qpsk_costas_bw', 0.005)

    @property
    def sstv_wsr(self) -> Union[int, float]:
        return self.config.get('sstv_wsr', 16000)

    @property
    def sstv_sync(self) -> bool:
        return self.config.get('sstv_sync', True)


class Satellite(gr.gr.hier_block2):
    @staticmethod
    def _validate_config(config: Mapping) -> bool:
        return (
            all(map(lambda x: x in config,
                    [
                        'name',
                        # 'enabled',        # optional
                        # 'min_elevation',  # optional
                        'frequencies',
                        # 'doppler',          # optional
                    ]))
            and config['frequencies']
        )

    def __init__(self,
                 config: Mapping,
                 sat_ephem_tle: Optional[tuple[ephem.EarthSatellite, tuple[str, str, str]]],
                 observer: Observer,
                 main_tune: Union[int, float],
                 samp_rate: Union[int, float],
                 output_directory: pathlib.Path,
                 executor):
        n = config.get('name', '')
        self.prefix = f'{self.__class__.__name__}{n and f": {n}"}'
        self.log = logging.getLogger(self.prefix)

        if not self._validate_config(config):
            raise ValueError(f'{self.prefix}: Invalid config!')

        super(Satellite, self).__init__(
            self.prefix,
            gr.gr.io_signature(1, 1, gr.gr.sizeof_gr_complex),
            gr.gr.io_signature(0, 0, 0)
        )

        self.sat_ephem_tle = sat_ephem_tle
        self.observer = observer
        self.executor = executor
        self.config = config
        self.output_directory = output_directory / self.name
        self.output_directory.mkdir(parents=True, exist_ok=True)
        self.events: list[Optional[utils.Event]] = [None, None, None]
        self.recorders = []

        for cfg in self.frequencies:
            try:
                r = SatRecorder(self, cfg, main_tune, samp_rate)
                if r.enabled:
                    self.connect(self, r)
                    self.recorders.append(r)
            except ValueError as e:
                self.log.warning('Skip freq `%s`: %s', cfg.get('freq', 'Unknown'), e)

    @property
    def is_runned(self) -> bool:
        return any(r.is_runned for r in self.recorders)

    def start(self):
        if self.enabled and not self.is_runned:
            self.log.info('START doppler=%s mode=%s decode=%s',
                          self.doppler,
                          [r.mode.value for r in self.recorders],
                          [r.decode.value for r in self.recorders])
            self.output_directory.mkdir(parents=True, exist_ok=True)
            self.start_event = None

            for r in self.recorders:
                r.decoder.start()
                r.radio.set_enabled(1)

    def stop(self):
        if self.is_runned:
            self.log.info('STOP')
            self.start_event = self.stop_event = None
            fin_key = sha256((self.name + str(dt.datetime.now())).encode()).hexdigest()

            for r in self.recorders:
                if r.is_runned:
                    r.radio.set_enabled(0)
                    r.decoder.finalize(self.executor, fin_key)

    def correct_doppler(self, observer: ephem.Observer):
        if self.is_runned and self.doppler:
            self.sat_ephem_tle[0].compute(observer)

            for r in self.recorders:
                if r.is_runned:
                    r.set_freq_offset(utils.doppler_shift(r.frequency, self.sat_ephem_tle[0].range_velocity))

    @property
    def name(self) -> str:
        return self.config['name']

    @property
    def enabled(self) -> bool:
        return self.config.get('enabled', True) and self.recorders

    @property
    def min_elevation(self) -> Union[int, float]:
        return self.config.get('min_elevation', 0.0)

    @property
    def frequencies(self) -> Mapping:
        return self.config['frequencies']

    @property
    def doppler(self) -> bool:
        return self.config.get('doppler', True)

    @property
    def start_event(self) -> utils.Event:
        return self.events[0]

    @start_event.setter
    def start_event(self, val: utils.Event):
        self.events[0] = val

    @property
    def stop_event(self) -> utils.Event:
        return self.events[1]

    @stop_event.setter
    def stop_event(self, val: utils.Event):
        self.events[1] = val

    @property
    def recalc_event(self) -> utils.Event:
        return self.events[2]

    @recalc_event.setter
    def recalc_event(self, val: utils.Event):
        self.events[2] = val



import datetime as dt
import logging
import pathlib
import shutil
import urllib.error
import urllib.parse
import urllib.request

from typing import Mapping, Optional, Union

import ephem

from sats_receiver import TLEDIR


class Tle:
    def __init__(self, config: Mapping):
        self.prefix = self.__class__.__name__
        self.log = logging.getLogger(self.prefix)

        self.config = {}
        self.tle_file = pathlib.Path(TLEDIR / 'dummy')
        self.last_update_tle = dt.datetime.fromtimestamp(0, dt.timezone.utc)
        self.objects: dict[str, tuple[ephem.EarthSatellite, tuple[str, str, str]]] = {}

        if not self.update_config(config):
            raise ValueError(f'{self.prefix}: Invalid config!')

        self.t_next = self.last_update_tle + dt.timedelta(days=self.update_period)

    def fill_objects(self):
        self.objects.clear()
        with self.tle_file.open() as f:
            for line in f:
                names = set()
                while len(line) <= 69:
                    names.add(line.strip())
                    line = f.readline()
                names.add(int(line[2:7]))

                l1 = line.rstrip()
                l2 = f.readline().rstrip()
                for name in names:
                    self.objects[name] = ephem.readtle(str(name), l1, l2), (str(name), l1, l2)

    def fetch_tle(self):
        try:
            urllib.request.urlretrieve(self.url, self.tle_file)
        except urllib.error.HTTPError as e:
            msg = f'Tle not fetched: {e}'
            if e.code == 400:
                msg = f'{msg}: "{e.url}"'
            self.log.error('%s', msg)
            return
        except (urllib.error.URLError, ValueError) as e:
            self.log.error('Tle not fetched: %s', e)
            return

        self.last_update_tle = dt.datetime.now(dt.timezone.utc)
        self.fill_objects()

        self.log.info('Tle updated')

        return 1

    def update_config(self, config: Mapping):
        """
        :return: True if config update success
        """

        if config != self.config:
            if not self._validate_config(config):
                self.log.warning('invalid new config!')
                return

            self.log.debug('reconf')
            self.config = config

            fn = pathlib.Path(urllib.parse.urlparse(self.url).path).name
            self.tle_file = pathlib.Path(TLEDIR / fn)
            if self.tle_file.is_file():
                self.last_update_tle = dt.datetime.fromtimestamp(self.tle_file.stat().st_mtime, dt.timezone.utc)
            else:
                if self.tle_file.is_dir():
                    shutil.rmtree(self.tle_file, True)
                else:
                    self.tle_file.unlink(True)
                self.tle_file.touch()
                self.last_update_tle = dt.datetime.fromtimestamp(0, dt.timezone.utc)

            self.fill_objects()

            return 1

    @staticmethod
    def _validate_config(config: Mapping) -> bool:
        return all(map(lambda x: x in config, [
            'url',
            'update_period',
        ]))

    @property
    def url(self) -> str:
        return self.config['url']

    @property
    def update_period(self) -> Union[int, float]:
        """
        Period of TLE update, days
        """

        return self.config['update_period']

    def action(self, t: dt.datetime):
        if t >= self.t_next and self.fetch_tle():
            self.t_next = self.last_update_tle + dt.timedelta(days=self.update_period)
            return 1

    def get(self, name: str) -> Optional[tuple[ephem.EarthSatellite, tuple[str, str, str]]]:
        """
        Get TLE info by satellite name or NORAD number

        :return: Tuple of EarthSatellite object and 3 lines of TLE. Or None
        """

        return self.objects.get(name, None)

    def get_ephem(self, name: str) -> Optional[ephem.EarthSatellite]:
        """
        Get TLE object by satellite name or NORAD number
        """

        x = self.objects.get(name, None)
        return x and x[0]

    def get_tle(self, name: str) -> Optional[tuple[str, str, str]]:
        """
        Get raw TLE lines by satellite name or NORAD number
        """

        x = self.objects.get(name, None)
        return x and x[1]
